from __future__ import absolute_import, unicode_literals


class RemovedInWagtailMenus26Warning(DeprecationWarning):
    pass


removed_in_next_version_warning = RemovedInWagtailMenus26Warning


class RemovedInWagtailMenus27Warning(PendingDeprecationWarning):
    pass


class RemovedInWagtailMenus28Warning(PendingDeprecationWarning):
    pass
